disclaimer = """
\033[91m
====================
     DISCLAIMER
====================
\033[0m
ESM_CLEANUP is a tool designed to reduce space blocked by simulations produced using ESM-Tools.
Before continuing, please be aware that automatically removing data from your experiment folders
includes the risk of loosing information that you might need in the future. Though some things
might be relatively safe to remove, other (like restart files) might not.
By continuing, you accept that neither the ESM-Tools developers nor anyone else but yourself 
take the responsibility of what happens to your data. 
"""

import os
import sys
import shutil
import questionary
import yaml
import time

per_model_folders = ["restart", "outdata", "input", "forcing", "bin", "config", "log"]


def is_experiment_folder(checkpath):
    all_files = os.listdir(checkpath)
    if (
        "scripts" in os.listdir(checkpath)
        and "bin" in os.listdir(checkpath)
        and "restart" in os.listdir(checkpath)
    ):
        return True
    return False


def pick_experiment_folder(folder):
    all_files = os.listdir(folder)
    all_with_size = [add_size_information(folder, dirname) for dirname in all_files] + [
        "The current folder is already the experiment folder I want to clean up."
    ]
    response = questionary.select(
        "Which experiment do you want to clean up?",
        choices=all_with_size,
    ).ask()
    if "The current folder" in response:
        return folder
    response_without_size = remove_size_information(response)
    return os.path.join(folder, response_without_size)


def pick_subfolder(folder):
    all_files = os.listdir(folder)
    all_with_size = [add_size_information(folder, dirname) for dirname in all_files]
    response = questionary.select(
        "Files from which folder do you want to remove?",
        choices=all_with_size,
    ).ask()
    response_without_size = remove_size_information(response)
    return response_without_size


def add_size_information(toplevel, item):
    full_filepath = os.path.join(toplevel, item)
    size, unit = format_size(inspect_size(full_filepath))
    return f"{item} ({size:.2f} {unit})"


def remove_size_information(name_with_file):
    name, size, unit = name_with_file.split()
    return name


def print_disclaimer():
    os.system("clear")
    print(disclaimer)
    response = questionary.select(
        f"Sure that you want to continue?",
        choices=[
            "Hell no, let me get out of here.",
            "Yes, I know what I am doing, and will be responsible for the outcome.",
        ],
    ).ask()
    if "Hell no" in response:
        sys.exit(0)


def read_in_yaml_file(filename):
    with open(filename, "r") as stream:
        runscript_dict = yaml.load(stream, Loader=yaml.FullLoader)

    print(runscript_dict)
    return runscript_dict


def assert_question(question):
    response = questionary.select(
        question,
        choices=[
            "No, there is something wrong.",
            "Yes, that is exactly what I want.",
            "I want to quit.",
        ],
    ).ask()
    if "quit" in response:
        sys.exit(0)
    if "Yes" in response:
        return True
    return False


def check_if_folder_exists(folder):
    if not os.path.isdir(folder):
        print(f"Folder not found: {folder}")
        sys.exit(-1)
    return


def print_folder_content(checkpath, saved_space):
    os.system("clear")
    all_files = os.listdir(checkpath)
    print(f"Files in folder {checkpath}:")
    print("=" * len(f"Files in folder {checkpath}:"))
    print()
    for thisfile in sorted(all_files):
        with_size = add_size_information(checkpath, thisfile)
        print("-- " + with_size)
    print()
    space_size, space_unit = format_size(saved_space)
    print(f"(Space saved so far in this session: {space_size:.2f} {space_unit})")
    print()


def file_size(somepath):
    return os.path.getsize(somepath)


def dir_size(somepath):
    size = 0
    for path, subdirs, files in os.walk(somepath):
        for somefile in files:
            full_filepath = os.path.join(path, somefile)
            if not os.path.islink(full_filepath):
                size += os.path.getsize(full_filepath)
    return size


def inspect_size(thisfile):
    if os.path.isdir(thisfile):
        total_size = dir_size(thisfile)
    else:
        total_size = file_size(thisfile)
    return total_size


def format_size(total_size):
    unit = "B"
    if total_size >= 1024:
        total_size = total_size / 1024.0
        unit = "kB"
    if total_size >= 1024:
        total_size = total_size / 1024.0
        unit = "MB"
    if total_size >= 1024:
        total_size = total_size / 1024.0
        unit = "GB"
    if total_size >= 1024:
        total_size = total_size / 1024.0
        unit = "TB"
    return total_size, unit


def ask_for_action(folder):
    question = "What do you want to do?"
    choices = ["Remove a WHOLE subfolder", "Quit"]

    subfolders = sorted(os.listdir(folder))

    if "restart" in subfolders:
        choices.insert(0, "Remove some of the files in a folder")
    run_folders = [
        subfolder for subfolder in subfolders if subfolder.startswith("run_")
    ]
    if len(run_folders) > 0:
        choices.insert(0, "Remove all run_* subfolders (recommended)")

    post_folders = [
        subfolder for subfolder in subfolders if subfolder.startswith("post_")
    ]
    if len(post_folders) > 0:
        choices.insert(0, "Remove all post_* subfolders (recommended)")

    response = questionary.select(
        question,
        choices=choices,
    ).ask()

    if "WHOLE" in response:
        return "remove_subfolder"
    elif "run_* subfolders" in response:
        return "remove_run_subfolders"
    elif "post_* subfolders" in response:
        return "remove_post_subfolders"
    elif "some of the files" in response:
        return "remove_some_files"

    return "quit"


def remove_subfolder(folder, saved_space):
    permitted_folders = ["restart", "scripts", "bin", "outdata"]
    choices = sorted(os.listdir(folder)) + ["(You know what, forget it.)"]
    choices = [choice for choice in choices if not choice in permitted_folders]
    question = "Pick a subfolder to remove:"

    response = questionary.select(question, choices=choices).ask()

    if "You know what" in response:
        sys.exit(0)

    if "restart" in response or "bin" in response:
        print(
            f"Folder {response} should not be removed entirely, and esm_cleanup won't do it."
        )

    full_path = os.path.join(folder, response)
    if not assert_question(
        f"Do you really want to remove the whole subfolder {full_path}?"
    ):
        return saved_space

    if os.path.isdir(full_path):
        size = inspect_size(full_path)
        saved_space += size
        shutil.rmtree(full_path, ignore_errors=True)
    return saved_space


def remove_run_subfolders(folder, saved_space):
    if not assert_question(f"Do you really want to remove ALL run_* subfolders?"):
        return saved_space
    subfolders = sorted(os.listdir(folder))
    run_folders = [
        subfolder for subfolder in subfolders if subfolder.startswith("run_")
    ]
    for run_folder in run_folders:
        full_path = os.path.join(folder, run_folder)
        if os.path.isdir(full_path):
            size = inspect_size(full_path)
            saved_space += size
            shutil.rmtree(full_path, ignore_errors=True)
    return saved_space


def remove_post_subfolders(folder, saved_space):
    if not assert_question(f"Do you really want to remove ALL post_* subfolders?"):
        return saved_space
    subfolders = sorted(os.listdir(folder))
    post_folders = [
        subfolder for subfolder in subfolders if subfolder.startswith("post_")
    ]
    for post_folder in post_folders:
        full_path = os.path.join(folder, post_folder)
        if os.path.isdir(full_path):
            size = inspect_size(full_path)
            saved_space += size
            shutil.rmtree(full_path, ignore_errors=True)
    return saved_space


def remove_some_files(folder, saved_space):
    subfolder = pick_subfolder(folder)
    basedir = folder + "/" + subfolder

    if subfolder in per_model_folders:
        models = os.listdir(basedir)
        if models == []:
            print("Folder seems to be empty.")
            time.wait(2)
            return saved_space
        models_with_sizes = [add_size_information(basedir, model) for model in models]
        model_with_size = questionary.select(
            "Restart files from which subfolder shall be removed?",
            choices=models_with_sizes,
        ).ask()

        model = remove_size_information(model_with_size)
        basedir = basedir + "/" + model

    contents = sorted(os.listdir(basedir))
    if contents == []:
        print()
        print("Folder seems to be empty.")
        time.sleep(2)
        return saved_space
    content_with_size = [add_size_information(basedir, content) for content in contents]
    answers = questionary.checkbox(
        "Select files to remove:",
        choices=content_with_size,
    ).ask()

    answers = [remove_size_information(answer) for answer in answers]

    if answers == []:
        return saved_space
    print(f"You have chosen to remove the following files:")
    for filename in answers:
        print(f" ---  {filename}")
    if not assert_question(f"Do you really want to remove these files?"):
        return saved_space

    for filename in answers:
        full_path = basedir + "/" + filename
        if os.path.isdir(full_path):
            size = inspect_size(full_path)
            saved_space += size
            shutil.rmtree(full_path, ignore_errors=True)
        elif os.path.isfile(full_path):
            size = inspect_size(full_path)
            saved_space += size
            os.remove(full_path)

    return saved_space



class yac:
    """

    Generates the configuration file for YAC coupler.
    """

    def __init__(
        self,
        full_config,
        nb_of_couplings=1,
        coupled_models=["echam", "fesom"],
        grids=["atmo", "feom"],
        runtime=1,
    ):

        self.name = "yac"

        self.namcouple = ['<?xml version="1.0" encoding="UTF-8"?>']
        self.namcouple += [
            "<!-- This coupling.xml was automatically generated by the esm-tools (Python) -->"
        ]
        self.namcouple += ["<coupling"]
        self.namcouple += ['\txmlns="http://www.w3schools.com"']
        self.namcouple += ['\txmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"']
        self.namcouple += [
            '\txsi:schemaLocation="http://www.w3schools.com coupling.xsd">'
        ]
        self.namcouple += [
            '\t<redirect redirect_of_root="'
            + str(full_config[self.name]["redirect_of_root"]).lower()
            + '" redirect_stdout="'
            + str(full_config[self.name]["redirect_stdout"]).lower()
            + '"/>'
        ]
        self.namcouple += ["\t<components>"]
        component_id = 1
        for component in full_config[self.name]["component_names"]:
            self.namcouple += ['\t\t<component id="' + str(component_id) + '">']
            self.namcouple += ["\t\t\t<name>" + str(component) + "</name>"]
            self.namcouple += ["\t\t\t<model>" + str(component) + "</model>"]
            self.namcouple += [
                "\t\t\t<simulated>"
                + str(full_config[coupled_models[component_id - 1]]["type"])
                + "</simulated>"
            ]
            self.namcouple += ["\t\t\t<transient_grid_refs>"]
            for field in range(nb_of_couplings):
                self.namcouple += [
                    '\t\t\t\t<transient_grid_ref id="'
                    + str(field + 1)
                    + '" transient_ref="'
                    + str(field + 1)
                    + '" grid_ref="'
                    + str(component_id)
                    + '" collection_size="1"/>'
                ]
            self.namcouple += ["\t\t\t</transient_grid_refs>"]
            self.namcouple += ["\t\t</component>"]
            component_id += 1
        self.namcouple += ["\t</components>"]
        self.namcouple += ["\t<transients>"]
        for field in range(nb_of_couplings):
            self.namcouple += [
                '\t\t<transient id="'
                + str(field + 1)
                + '" transient_standard_name="??transient_name'
                + str(field + 1)
                + '??"/>'
            ]  # Placeholder ??..?? will be replace further down.
        self.namcouple += ["\t</transients>"]
        self.namcouple += ["\t<grids>"]
        grid_id = 1
        for grid in grids:
            self.namcouple += [
                '\t\t<grid id="' + str(grid_id) + '" alias_name="' + str(grid) + '" />'
            ]
            grid_id += 1
        self.namcouple += ["\t</grids>"]
        self.namcouple += ["\t<dates>"]
        self.namcouple += [
            "\t\t\t<start_date>"
            + str(full_config["general"]["initial_date"])
            + "</start_date>"
        ]
        self.namcouple += [
            "\t\t\t<end_date>"
            + str(full_config["general"]["next_date"])
            + "</end_date>"
        ]
        self.namcouple += [
            "\t\t\t<calendar>" + str(full_config[self.name]["calendar"]) + "</calendar>"
        ]
        self.namcouple += ["\t</dates>"]
        self.namcouple += [
            "\t<timestep_unit>"
            + str(full_config[self.name]["coupling_time_step_unit"])
            + "</timestep_unit>"
        ]
        self.namcouple += ["\t<couples>"]
        self.namcouple += ["\t\t<couple>"]
        component_id = 1
        for component in coupled_models:
            self.namcouple += [
                "\t\t\t<component"
                + str(component_id)
                + ' component_id="'
                + str(component_id)
                + '" />'
            ]
            component_id += 1

        self.next_coupling = 1

    def add_coupling(self, field, transient_id, direction, config):
        import sys

        coupling_field = field.replace("<--", "%").replace("--", "&")
        source, rest = coupling_field.split("%")
        source = source.strip()
        interpolation, target = rest.split("&")
        target = target.strip()
        interpolation = interpolation.strip()

        # Replacing placeholder ??..?? with field keyname source
        matching = [
            s
            for s in self.namcouple
            if "??transient_name" + str(transient_id) + "??" in s
        ]
        index = self.namcouple.index(matching[0])
        self.namcouple[index] = self.namcouple[index].replace(
            "??transient_name" + str(transient_id) + "??", source
        )

        self.namcouple += [
            '\t\t\t\t<transient_couple transient_id="' + str(transient_id) + '">'
        ]
        self.namcouple += [
            '\t\t\t\t\t<source component_ref="'
            + str(config["coupling_directions"][direction]["source"]["component_id"])
            + '" transient_grid_ref="'
            + str(transient_id)
            + '" />'
        ]
        self.namcouple += [
            '\t\t\t\t\t<target transient_grid_ref="' + str(transient_id) + '" />'
        ]
        self.namcouple += ["\t\t\t\t\t<timestep>"]
        self.namcouple += [
            "\t\t\t\t\t\t<source>"
            + str(config["coupling_directions"][direction]["source"]["timestep"])
            + "</source>"
        ]
        self.namcouple += [
            "\t\t\t\t\t\t<target>"
            + str(config["coupling_directions"][direction]["target"]["timestep"])
            + "</target>"
        ]
        self.namcouple += [
            '\t\t\t\t\t\t<coupling_period operation="'
            + str(config["coupling_directions"][direction]["operation"])
            + '">'
            + str(config["coupling_time_step"])
            + "</coupling_period>"
        ]
        self.namcouple += [
            "\t\t\t\t\t\t<source_timelag>"
            + str(config["coupling_directions"][direction]["source"]["timelag"])
            + "</source_timelag>"
        ]
        self.namcouple += [
            "\t\t\t\t\t\t<target_timelag>"
            + str(config["coupling_directions"][direction]["target"]["timelag"])
            + "</target_timelag>"
        ]
        self.namcouple += ["\t\t\t\t\t</timestep>"]
        self.namcouple += [
            '\t\t\t\t\t<interpolation_requirements use_source_mask="'
            + str(
                config["coupling_directions"][direction]["source"]["use_mask"]
            ).lower()
            + '" use_target_mask="'
            + str(
                config["coupling_directions"][direction]["source"]["use_mask"]
            ).lower()
            + '">'
        ]

        line = '<interpolation method="' + str(interpolation) + '" '

        for option in config["interpolation_methods"][interpolation]["options"]:
            value = config["interpolation_methods"][interpolation]["options"][option]
            line += " " + str(option) + ' ="' + str(value) + '"'
        line += "/>"
        self.namcouple += ["\t\t\t\t\t\t\t" + line]

        self.namcouple += ["\t\t\t\t\t</interpolation_requirements>"]
        self.namcouple += [
            '\t\t\t\t\t<debug_mode at_source_before_interpolation="'
            + str(
                config["coupling_directions"][direction]["source"][
                    "debug_before_interpolation"
                ]
            ).lower()
            + '" at_source_after_interpolation="'
            + str(
                config["coupling_directions"][direction]["source"][
                    "debug_before_interpolation"
                ]
            ).lower()
            + '" at_target="'
            + str(config["coupling_directions"][direction]["target"]["debug"]).lower()
            + '"/>'
        ]
        self.namcouple += [
            "\t\t\t\t\t<enforce_write_restart>"
            + str(config["coupling_directions"][direction]["write_restart"]).lower()
            + "</enforce_write_restart>"
        ]
        self.namcouple += [
            '\t\t\t\t\t<enforce_write_weight_file filename="'
            + str(source)
            + '_weight">'
            + str(config["coupling_directions"][direction]["write_weight"]).lower()
            + "</enforce_write_weight_file>"
        ]
        self.namcouple += ["\t\t\t\t</transient_couple>"]

    def print_config_files(self):
        for line in self.namcouple:
            print(line)

    def add_output_file(self, lefts, rights, leftmodel, rightmodel, config):
        out_file = []

        coupling = self.next_coupling

        if self.next_coupling < 10:
            this_coupling = "0" + str(coupling)
        else:
            this_coupling = str(coupling)

        for lefty in lefts:
            out_file.append(lefty + "_" + leftmodel + "_" + this_coupling + ".nc")
        for righty in rights:
            out_file.append(righty + "_" + rightmodel + "_" + this_coupling + ".nc")

        self.next_coupling += 1

        if "outdata_files" not in config:
            config["outdata_files"] = {}
        if "outdata_in_work" not in config:
            config["outdata_in_work"] = {}
        if "outdata_sources" not in config:
            config["outdata_sources"] = {}

        for thisfile in out_file:

            config["outdata_files"][thisfile] = thisfile
            config["outdata_in_work"][thisfile] = thisfile
            config["outdata_sources"][thisfile] = thisfile

    def add_restart_files(self, restart_file, fconfig):
        config = fconfig[self.name]
        gconfig = fconfig["general"]
        # enddate = "_" + str(gconfig["end_date"].year) + str(gconfig["end_date"].month) + str(gconfig["end_date"].day)
        # parentdate = "_" + str(config["parent_date"].year) + str(config["parent_date"].month) + str(config["parent_date"].day)
        enddate = "_" + gconfig["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        parentdate = "_" + config["parent_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )

        if "restart_out_files" not in config:
            config["restart_out_files"] = {}
        if "restart_out_in_work" not in config:
            config["restart_out_in_work"] = {}
        if "restart_out_sources" not in config:
            config["restart_out_sources"] = {}

        if "restart_in_files" not in config:
            config["restart_in_files"] = {}
        if "restart_in_in_work" not in config:
            config["restart_in_in_work"] = {}
        if "restart_in_sources" not in config:
            config["restart_in_sources"] = {}

        config["restart_out_files"][restart_file] = restart_file
        config["restart_out_files"][restart_file + "_recv"] = restart_file + "_recv"

        config["restart_out_in_work"][restart_file] = restart_file  # + enddate
        config["restart_out_in_work"][restart_file + "_recv"] = (
            restart_file + "_recv"
        )  # + enddate

        config["restart_out_sources"][restart_file] = restart_file
        config["restart_out_sources"][restart_file + "_recv"] = restart_file + "_recv"

        config["restart_in_files"][restart_file] = restart_file
        config["restart_in_in_work"][restart_file] = restart_file
        if restart_file not in config["restart_in_sources"]:
            config["restart_in_sources"][restart_file] = restart_file

    def prepare_restarts(self, restart_file, all_fields, model, config):
        enddate = "_" + config["general"]["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        # enddate = "_" + str(config["general"]["end_date"].year) + str(config["general"]["end_date"].month) + str(config["general"]["end_date"].day)
        import glob
        import os
        import subprocess

        print("Preparing YAC restart files from initial run...")
        exe = config[model]["executable"]
        print(restart_file, all_fields, model, exe)
        cwd = os.getcwd()
        os.chdir(config["general"]["thisrun_work_dir"])
        filelist = ""
        for field in all_fields:
            print(field + "-" + model)
            thesefiles = glob.glob(field + "_" + exe + "_*.nc")
            print(thesefiles)
            for thisfile in thesefiles:
                print("cdo showtime " + thisfile + " 2>/dev/null | wc -w")
                lasttimestep = (
                    subprocess.check_output(
                        "cdo showtime " + thisfile + " 2>/dev/null | wc -w", shell=True
                    )
                    .decode("utf-8")
                    .rstrip()
                )
                # print (lasttimestep)

                print(
                    "cdo -O seltimestep,"
                    + str(lasttimestep)
                    + " "
                    + thisfile
                    + " onlyonetimestep.nc"
                )
                os.system(
                    "cdo -O seltimestep,"
                    + str(lasttimestep)
                    + " "
                    + thisfile
                    + " onlyonetimestep.nc"
                )
                print("ncwa -O -a time onlyonetimestep.nc notimestep_" + field + ".nc")
                os.system(
                    "ncwa -O -a time onlyonetimestep.nc notimestep_" + field + ".nc"
                )
                filelist += "notimestep_" + field + ".nc "
                print(filelist)
        print("cdo merge " + filelist + " " + restart_file)  # + enddate)
        os.system("cdo merge " + filelist + " " + restart_file)  # + enddate)
        rmlist = glob.glob("notimestep*")
        rmlist.append("onlyonetimestep.nc")
        for rmfile in rmlist:
            print("rm " + rmfile)
            os.system("rm " + rmfile)
        os.chdir(cwd)

    def finalize(self, destination_dir):
        self.namcouple += ["\t\t</couple>"]
        self.namcouple += ["\t</couples>"]
        self.namcouple += ["</coupling>"]
        endline = ""
        with open(destination_dir + "/coupling.xml", "w") as namcouple:
            for line in self.namcouple:
                namcouple.write(endline)
                namcouple.write(line)
                endline = "\n"



def write_hostfile(config):
    path = "hostfile_srun"
    if "multi_srun" in config["general"]:
        for run_type in list(config["general"]["multi_srun"]):
            current_hostfile = path + "_" + run_type
            write_one_hostfile(current_hostfile, config)
    else:
        write_one_hostfile(path, config)
    return config


def write_one_hostfile(hostfile, config):
    mpirun_options = ""

    for model in config["general"]["valid_model_names"]:
        end_core = config[model].get("end_proc", None)
        start_proc = config[model].get("start_proc", None)

        no_cpus = end_proc - start_proc

        if "execution_command" in config[model]:
            mpirun_options += (
                " -np " + no_cpus + " ./" + config[model]["execution_command"] + " :"
            )
        elif "executable" in config[model]:
            mpirun_options += (
                " -np " + no_cpus + " ./" + config[model]["executable"] + " :"
            )

    mpirun_options = mpirun_options[:-1]  # remove trailing ":"

    with open(hostfile, "w") as hostfile:
        hostfile.write(mpirun_options)
    return config



import argparse
import os

from .esm_cleanup import *


def evaluate_arguments():

    """The arg parser for interactive use"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-r", "--runscript", help="Name of the experiment's runscript", default=None
    )

    parser.add_argument(
        "-f",
        "--folder",
        help="Location to experiment's toplevel folder",
        default=".",
    )

    parser.add_argument(
        "-e",
        "--expid",
        help="Experiment name",
        default=None,
    )

    args = vars(parser.parse_args())

    runscript = None
    folder = "."

    if "runscript" in args and args["runscript"] and "expid" in args and args["expid"]:
        runscript_dict = read_in_yaml_file(args["runscript"])
        folder = runscript_dict["general"]["base_dir"] + "/" + args["expid"]
    elif "folder" in args:
        folder = args["folder"]
    else:
        print(
            "Either specify runscript and expid, or experiment folder to identify experiment to clean up."
        )
        sys.exit(-1)

    if folder == ".":
        folder = os.getcwd()

    return folder


def main_loop(folder):
    saved_space = 0
    while True:
        print_folder_content(folder, saved_space)
        action = ask_for_action(folder)

        if action == "remove_subfolder":
            saved_space = remove_subfolder(folder, saved_space)
        elif action == "remove_run_subfolders":
            saved_space = remove_run_subfolders(folder, saved_space)
        elif action == "remove_some_files":
            saved_space = remove_some_files(folder, saved_space)
        else:
            sys.exit(0)


def main():

    print_disclaimer()
    orgfolder = evaluate_arguments()

    while True:
        folder = orgfolder
        while not is_experiment_folder(folder):
            answer = pick_experiment_folder(folder)
            if answer == folder:
                break
            folder = answer

        if assert_question(f"Continue to work on folder {folder}?"):
            break

    check_if_folder_exists(folder)
    main_loop(folder)



"""
``esm-runscripts`` Core Plugins for dealing with Fortran Namelists.

Provides plugins for loading, modifying, deleting, and writing Fortran
Namelists as part of the ``esm-runscripts`` recipe. All plugins are found under
the class Namelist as static methods. A deprecated class ``namelist`` (small "n") is
provided, which warns you when it is used.
"""
import logging
import os
import sys
import warnings

import f90nml

from esm_parser import user_error


class Namelist:
    """Methods for dealing with FORTRAN namelists"""

    @staticmethod
    def nmls_load(mconfig):
        """
        Loads Fortran namelists into the configuration dictionary.

        User Information
        ----------------
        To associate namelists with a specific model, you should have a section
        in your configuration that lists the namelists::

            fesom:
                namelists:
                    - "namelist.config"
                    - "namelist.oce"
                    - "namelist.ice"
                    - "namelist.diag"

        Programmer Information
        ----------------------
        The namelists are represented by f90nml Namelist objects, and are
        stored under::

            mconfig["namelists"]["namelist.echam"]``

        This would point to the ECHAM namelist as a f90nml object, which
        closely resembles a dictionary.

        The actual namelists to load are listed in the raw configuration as a
        list of strings::

            mconfig['namelists'] = ['nml1', 'nml2', 'nml3', ...]

        Namelists are assumed to have been copied to
        ``mconfig["thisrun_config_dir"]``, and are loaded from there.

        If the ``mconfig`` has a key ``"namelist_case"`` equal to "uppercase",
        the uppercase attribute of the f90nml representation of the namelist is
        set to ``True``.

        Parameters
        ----------
        mconfig : dict
            The model (e.g. ECHAM, FESOM, NEMO or OIFS) configuration

        Returns
        -------
        mconfig : dict
            The modified configuration.
        """
        nmls = mconfig.get("namelists", [])
        mconfig["namelists"] = dict.fromkeys(nmls)
        for nml in nmls:
            if os.path.isfile(os.path.join(mconfig["thisrun_config_dir"], nml)):
                logging.debug("Loading %s", nml)
                try:
                    mconfig["namelists"][nml] = f90nml.read(
                        os.path.join(mconfig["thisrun_config_dir"], nml)
                    )
                except (StopIteration, ValueError) as e:
                    user_error(
                        "Namelist format",
                        f"The namelist ``{nml}`` has a formatting issue. Please, "
                        f"fix it before proceeding.\n\n``Hint:`` {e}",
                        dsymbols=["``", "'"],
                    )
            else:
                mconfig["namelists"][nml] = f90nml.namelist.Namelist()
            if mconfig.get("namelist_case") == "uppercase":
                mconfig["namelists"][nml].uppercase = True
        return mconfig

    @staticmethod
    def nmls_remove(mconfig):
        """
        Removes an element from a namelist chapter.

        User Information
        ----------------
        In the configuration file, assume you have::

            echam:
                namelist_changes:
                    namelist.echam:
                        radctl:
                            co2vmr: "remove_from_namelist"

        In this case, the entry co2vmr would be deleted from the radctl section
        of namelist.echam.

        Programmer Information
        ----------------------
        IDEA(PG): Maybe we can provide examples of how these functions are used
        in the code?

        Parameters
        ----------
        mconfig : dict
            The model (e.g. ECHAM, FESOM, NEMO or OIFS) configuration

        Returns
        -------
        mconfig : dict
            The modified configuration.
        """

        namelist_changes = mconfig.get("namelist_changes", {})
        namelist_removes = []
        for namelist in list(namelist_changes):
            changes = namelist_changes[namelist]
            logging.debug("Determining remove entires for %s", namelist)
            logging.debug("All changes: %s", changes)
            for change_chapter in list(changes):
                change_entries = changes[change_chapter]
                for key in list(change_entries):
                    value = change_entries[key]
                    if value == "remove_from_namelist":
                        namelist_removes.append((namelist, change_chapter, key))

                        # the key is probably coming from esm_tools config
                        # files or from a user runscript. It can contain lower
                        # case, but the original Fortran namelist could be in
                        # any case combination. Here `original_key` is coming
                        # from the default namelist and may contain mixed case.
                        # `key` is the processed variable from f90nml module and
                        # is lowercase.
                        remove_original_key = False

                        # traverse the namelist chapter and see if a mixed case
                        # variable is also found
                        for key2 in namelist_changes[namelist][change_chapter]:
                            # take care of the MiXeD FORTRAN CaSeS
                            if key2.lower() == key.lower() and key2 != key:
                                original_key = key2
                                remove_original_key = True
                                namelist_removes.append(
                                    (namelist, change_chapter, original_key)
                                )

                        # remove both lowercase and mixed case variables
                        del namelist_changes[namelist][change_chapter][key]
                        if remove_original_key:
                            del namelist_changes[namelist][change_chapter][original_key]

                        # mconfig instead of config, Grrrrr
                        print(
                            f"- NOTE: removing the variable: {key} from the namelist: {namelist}"
                        )

        for remove in namelist_removes:
            namelist, change_chapter, key = remove
            logging.debug("Removing from %s: %s, %s", namelist, change_chapter, key)
            if key in mconfig["namelists"][namelist].get(change_chapter, {}):
                del mconfig["namelists"][namelist][change_chapter][key]
            elif "%" in key:
                namvar, prop = key.split("%")
                del mconfig["namelists"][namelist][change_chapter][namvar][prop]
            else:
                logging.debug(
                    "Unable to remove %s: %s, %s", namelist, change_chapter, key
                )
        return mconfig

    @staticmethod
    def nmls_modify(mconfig):
        """
        Performs namelist changes.

        User Information
        ----------------
        In the configuration file, you should have a section as::

            echam:
                namelist_changes:
                    namelist.echam:
                        radctl:
                            co2vmr: 1200e-6


        This would change the value of the echam namelist (namelist.echam),
        subsection radctl, entry co2vmr to the value 1200e-6.

        Programmer Information
        ----------------------
        IDEA(PG): Maybe we can provide examples of how these functions are used
        in the code?

        Note
        ----
        Actual changes are performed by the f90nml package patch fuction. See
        here: https://tinyurl.com/y4ydz363

        Parameters
        ----------
        mconfig : dict
            The model (e.g. ECHAM, FESOM, NEMO or OIFS) configuration

        Returns
        -------
        mconfig : dict
            The modified configuration.
        """
        namelist_changes = mconfig.get("namelist_changes", {})
        for namelist, changes in namelist_changes.items():
            mconfig["namelists"][namelist].patch(changes)
        return mconfig

    @staticmethod
    def echam_transient_forcing(config):
        """
        Allows for ECHAM specific transient orbital and greenhouse gas forcing .

        User Information
        ----------------
        In the configuration file, assume you have::

            echam:
                use_transient_forcing: True
                transient_forcing_table: /some/path/to/a/table

        In this case, the entries for co2vmr, n2ovmr, ch4vmr, cecc, cobl, clonp
        would be extracted from this table according to the current model year,
        and the radctl section of namelist.echam is modified.

        The forcing table should look like this::

            # Year; CO2; CH4; N2O; Eccentricity; Obliquity; Perihelion
            1850; 0.000187; 3.779100e-07; 2.064600e-07; 0.018994; 22.944859; 294.23880

        The esm-tools will first check if the current year exists, extract the
        relevant values for you, and put everything into the radctl section of
        ``namelist.echam``

        Parameters
        ----------
        config : dict
            The config

        Returns
        -------
        config : dict
            The modified configuration.
        """
        if "echam" in config["general"]["valid_model_names"]:
            # Get the echam namelist:
            nml = config["echam"]["namelists"]["namelist.echam"]
            # Get the current radtl chapter or make a new empty one:
            radctl = nml.get("radctl", f90nml.namelist.Namelist())
            if config["echam"].get("use_transient_forcing", False):
                import pandas as pd

                try:
                    forcing_table = pd.read_csv(
                        config["echam"]["transient_forcing_table"], sep=";", index_col=0
                    )
                    co2, n2o, ch4, cecc, cobld, clonp = forcing_table.loc[
                        config["general"]["current_date"].year
                    ]
                    radctl["co2vmr"] = co2
                    radctl["n2ovmr"] = n2o
                    radctl["ch4vmr"] = ch4
                    radctl["cecc"] = cecc
                    radctl["cobld"] = cobld
                    radctl["clonp"] = clonp
                    print(
                        "-------------------------------------------------------------"
                    )
                    print("")
                    print(
                        "              > Applying transient foricng in echam namelist!"
                    )
                    print("")
                    print(
                        "--------------------------------------------------------------"
                    )
                    print("             > The new values are:")
                    print(f"             CO2:  {radctl['co2vmr']}")
                    print(f"             N2O:  {radctl['n2ovmr']}")
                    print(f"             CH4:  {radctl['ch4vmr']}")
                    print(f"             CECC: {radctl['cecc']}")
                    print(f"             COBL: {radctl['cobld']}")
                    print(f"             CLONP:{radctl['clonp']}")
                except Exception as e:
                    # Haha something went wrong. Let's be polite about it though
                    print(
                        "There was a problem with reading in the forcing from the transient forcing table"
                    )
                    print()
                    print("Sorry")
                    print()
                    print("Please be sure to use the correct format of your table!")
                    print("It should be the following:")
                    print(
                        "# Model Year; CO2; N2O; CH4; Eccentricty; Obliquity; Perihelion"
                    )
                    print("Commented lines (with a #) will be ignored in that file")
                    print(
                        "Please note that you need to use a semicolon (;) as a seperator"
                    )
                    print()
                    print("Also, make sure that you set a valid filepath")
                    print("We were looking for the following:")
                    try:
                        print(config["echam"]["transient_forcing_table"])
                    except KeyError:
                        print(
                            "Oops, looks like you didn't specify which forcing table to use!"
                        )
                        print("You need to set in your echam configuration:")
                        print("echam:")
                        print("    transient_forcing_table: /path/to/your/table")
                        sys.exit(1)
                    sys.exit(1)
        return config

    @staticmethod
    def apply_echam_disturbance(config):
        """
        Applies a disturbance to the DYNCTL chapter of the echam namelist via the enstdif

        Relevant configuration entries:
        * disturbance_years (list of int): Which year to apply the disturbance
        * distrubance (float): Value to apply. Default can be found in echam.yaml
        """
        if "echam" in config["general"]["valid_model_names"]:
            # Get the echam namelist:
            nml = config["echam"]["namelists"]["namelist.echam"]
            # Get the current dynctl chapter or make a new empty one:
            dynctl = nml.get("dynctl", f90nml.namelist.Namelist())
            # Determine which years the user wants to have disturbed:
            if os.path.isfile(
                config["general"]["experiment_scripts_dir"] + "/disturb_years.dat"
            ):
                with open(
                    config["general"]["experiment_scripts_dir"] + "/disturb_years.dat"
                ) as f:
                    disturbance_file = [
                        int(line.strip()) for line in f.readlines() if line.strip()
                    ]
                if config["general"]["verbose"]:
                    print(disturbance_file)
            else:
                disturbance_file = None
                if config["general"]["verbose"]:
                    print(
                        "WARNING: "
                        + config["general"]["experiment_scripts_dir"]
                        + "/disturb_years.dat",
                        "was not found",
                    )
            disturbance_years = disturbance_file or config["echam"].get(
                "disturbance_years", []
            )
            current_year = config["general"]["current_date"].year
            if current_year in disturbance_years:
                print("-------------------------------------------------------")
                print("")
                print("              > Applying disturbance in echam namelist!")
                print("")
                print("-------------------------------------------------------")
                dynctl["enstdif"] = config["echam"].get("disturbance", 1.000001)
                nml["dynctl"] = dynctl
            else:
                if config["general"]["verbose"]:
                    print("Not applying disturbance in echam namelist.")
                    print(
                        "Current year",
                        current_year,
                        "disturbance_years",
                        disturbance_years,
                    )
        return config

    @staticmethod
    def nmls_finalize(mconfig, verbose):
        """
        Writes namelists to disk after all modifications have finished.

        User Information
        ----------------
        Part of the main log output will be a section specifing the actual
        namelists that have been used for your simulation, including all
        relevant additions, removals, or changes.

        Programmer Information
        ----------------------
        A copy of the f90nml object representations of the namelists is stored
        under the dictionary key "namelist_objs", as a dictionary of
        ("namelist_name", f90nml_objfect) key/value pairs.

        Warning
        -------
        Removing this step from your recipe might result in a broken run,
        as the namelists will not be present in their desired form! Even if
        your model runs, it might not contain all user-required changes.

        Parameters
        ----------
        mconfig : dict
            The model (e.g. ECHAM, FESOM, NEMO or OIFS) configuration

        Returns
        -------
        mconfig : dict
            The modified configuration.
        """
        all_nmls = {}

        for nml_name, nml_obj in mconfig.get("namelists", {}).items():
            with open(
                os.path.join(mconfig["thisrun_config_dir"], nml_name), "w"
            ) as nml_file:
                nml_obj.write(nml_file)
            all_nmls[nml_name] = nml_obj  # PG: or a string representation?
        mconfig["namelist_objs"] = all_nmls
        if verbose:
            mconfig = Namelist.nmls_output(mconfig)
        return mconfig

    @staticmethod
    def nmls_output(mconfig):
        all_nmls = {}

        for nml_name, nml_obj in mconfig.get("namelists", {}).items():
            all_nmls[nml_name] = nml_obj  # PG: or a string representation?
        for nml_name, nml in all_nmls.items():
            message = f"\nFinal Contents of {nml_name}:"
            print(message)
            print(len(message) * "-")
            nml.write(sys.stdout)
            print("-" * 80)
            print(f"::: end of the contents of {nml_name}\n")
        return mconfig

    @staticmethod
    def nmls_output_all(config):
        print(
            "\n" "- Namelists modified according to experiment specifications..."
        )
        for model in config["general"]["valid_model_names"]:
            config[model] = nmls_output(config[model], config["general"]["verbose"])
        return config

    @staticmethod
    def apply_iceberg_calving(config):
        """
        Calculates new number of icebergs when icesheet coupling is turned on

        Relevant configuration entries:
        """
        if "fesom" in config["general"]["valid_model_names"] and config["fesom"].get(
            "use_icebergs", False
        ):
            # Get the fesom config namelist:
            nml = config["fesom"]["namelists"]["namelist.config"]
            # Get the current icebergs chapter or make a new empty one:
            icebergs = nml.get("icebergs", f90nml.namelist.Namelist())
            # Determine if icesheet coupling is enabled:
            if config["fesom"].get("use_icesheet_coupling", False):
                icebergs["use_icesheet_coupling"] = True
                if os.path.isfile(
                    config["general"]["experiment_couple_dir"] + "/num_non_melted_icb_file"
                ):
                    with open(
                        config["general"]["experiment_couple_dir"] + "/num_non_melted_icb_file"
                    ) as f:
                        ib_num_old = [
                            int(line.strip()) for line in f.readlines() if line.strip()
                        ][0]
                elif config["general"].get("chunk_number", 0) == 1:
                    ib_num_old = 0
                else:
                    print("Something went wrong! Continue without old icebergs.")
                    ib_num_old = 0

                print(" * iceberg_dir = ", config["fesom"].get("iceberg_dir"))
                ib_num_new = sum(1 for line in open(config["fesom"].get("iceberg_dir") + "/LON.dat"))
                icebergs["ib_num"] = ib_num_old + ib_num_new
                nml["icebergs"] = icebergs
        return config


class namelist(Namelist):
    """Legacy class name. Please use Namelist instead!"""

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Please change your code to use Namelist!",
            DeprecationWarning,
            stacklevel=2,
        )

        super(namelist, self).__init__(*args, **kwargs)



def prepare_environment(config):
    environment_dict = {
            "VAR_IN_SHELL": "same_value",
            "SETUP_NAME": config["general"]["setup_name"],
            }
    return environment_dict
